import logging

import click

from bubuku.amazon import Amazon
from bubuku.config import load_config, KafkaProperties, Config
from bubuku.features.remote_exec import RemoteCommandExecutorCheck
from bubuku.id_generator import get_broker_id_policy
from bubuku.zookeeper import load_exhibitor_proxy, BukuExhibitor
from bubuku.zookeeper.exhibior import AWSExhibitorAddressProvider

_LOG = logging.getLogger('bubuku.cli')


def __get_opt_broker_id(broker_id: str, config: Config, zk: BukuExhibitor, amazon: Amazon) -> str:
    if not broker_id:
        kafka_properties = KafkaProperties(config.kafka_settings_template, '/tmp/tmp.props'.format(config.kafka_dir))
        broker_id_manager = get_broker_id_policy(config.id_policy, zk, kafka_properties, amazon)
        broker_id = broker_id_manager.get_broker_id()
        _LOG.info('Will use broker_id {}'.format(broker_id))
    running_brokers = zk.get_broker_ids()
    if broker_id not in running_brokers:
        raise Exception('Broker id {} is not registered ({}), can not restart'.format(broker_id, running_brokers))
    return broker_id


def __prepare_configs():
    config = load_config()
    _LOG.info('Using config: {}'.format(config))
    amazon = Amazon()
    zookeeper = load_exhibitor_proxy(AWSExhibitorAddressProvider(amazon, config.zk_stack_name), config.zk_prefix)
    return config, amazon, zookeeper


logging.basicConfig(level=getattr(logging, 'INFO', None))


@click.group()
def cli():
    pass


@cli.command('restart')
@click.option('--broker', type=click.STRING,
              help='Broker id to restart. By default current broker id is restarted')
def restart_broker(broker: str):
    config, amazon, zookeeper = __prepare_configs()
    broker_id = __get_opt_broker_id(broker, config, zookeeper, amazon)
    RemoteCommandExecutorCheck.register_restart(zookeeper, broker_id)


@cli.command('rebalance')
@click.option('--broker', type=click.STRING,
              help="Broker instance on which to perform rebalance. By default, any free broker will start it")
def rebalance_partitions(broker: str):
    config, amazon, zookeeper = __prepare_configs()
    broker_id = __get_opt_broker_id(broker, config, zookeeper, amazon) if broker else None
    RemoteCommandExecutorCheck.register_rebalance(zookeeper, broker_id)


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3
"""General Kafka Start Script."""

import logging

from bubuku import health
from bubuku.amazon import Amazon
from bubuku.broker import BrokerManager
from bubuku.config import load_config, KafkaProperties, Config
from bubuku.controller import Controller
from bubuku.features.rebalance import RebalanceOnStartCheck, RebalanceOnBrokerListChange
from bubuku.features.remote_exec import RemoteCommandExecutorCheck
from bubuku.features.restart_if_dead import CheckBrokerStopped
from bubuku.features.restart_on_zk_change import CheckExhibitorAddressChanged
from bubuku.features.swap_partitions import CheckBrokersDiskImbalance
from bubuku.features.terminate import register_terminate_on_interrupt
from bubuku.id_generator import get_broker_id_policy
from bubuku.utils import CmdHelper
from bubuku.zookeeper import BukuExhibitor, load_exhibitor_proxy
from bubuku.zookeeper.exhibior import AWSExhibitorAddressProvider

_LOG = logging.getLogger('bubuku.main')


def apply_features(features: dict, controller: Controller, buku_proxy: BukuExhibitor, broker: BrokerManager,
                   kafka_properties: KafkaProperties, amazon: Amazon) -> list:
    for feature, config in features.items():
        if feature == 'restart_on_exhibitor':
            controller.add_check(CheckExhibitorAddressChanged(buku_proxy, broker))
        elif feature == 'rebalance_on_start':
            controller.add_check(RebalanceOnStartCheck(buku_proxy, broker))
        elif feature == 'rebalance_on_brokers_change':
            controller.add_check(RebalanceOnBrokerListChange(buku_proxy, broker))
        elif feature == 'balance_data_size':
            controller.add_check(GenerateDataSizeStatistics(buku_proxy, broker, CmdHelper(),
                                                            kafka_properties.get_property("log.dirs").split(",")))
            controller.add_check(CheckBrokersDiskImbalance(buku_proxy, broker, config["diff_threshold_mb"] * 1024))
        elif feature == 'graceful_terminate':
            register_terminate_on_interrupt(controller, broker)
        elif feature == 'use_ip_address':
            kafka_properties.set_property('advertised.host.name', amazon.get_own_ip())
        else:
            _LOG.error('Using of unsupported feature "{}", skipping it'.format(feature))


def main():
    logging.basicConfig(level=getattr(logging, 'INFO', None))

    config = load_config()
    _LOG.info("Using configuration: {}".format(config))
    kafka_properties = KafkaProperties(config.kafka_settings_template,
                                       '{}/config/server.properties'.format(config.kafka_dir))

    amazon = Amazon()

    address_provider = AWSExhibitorAddressProvider(amazon, config.zk_stack_name)

    _LOG.info("Loading exhibitor configuration")
    buku_proxy = load_exhibitor_proxy(address_provider, config.zk_prefix)

    _LOG.info("Loading broker_id policy")
    broker_id_manager = get_broker_id_policy(config.id_policy, buku_proxy, kafka_properties, amazon)

    _LOG.info("Building broker manager")
    broker = BrokerManager(config.kafka_dir, buku_proxy, broker_id_manager, kafka_properties)

    _LOG.info("Creating controller")
    controller = Controller(broker, buku_proxy, amazon)

    controller.add_check(CheckBrokerStopped(broker, buku_proxy))
    controller.add_check(RemoteCommandExecutorCheck(buku_proxy, broker))
    apply_features(config.features, controller, buku_proxy, broker, kafka_properties, amazon)

    _LOG.info('Starting health server')
    health.start_server(config.health_port)

    _LOG.info('Starting main controller loop')
    controller.loop()


if __name__ == '__main__':
    main()



import json
import logging

import boto3
import requests

_LOG = logging.getLogger('bubuku.amazon')


class Amazon(object):
    NONE = object()

    def __init__(self):
        self.document = None
        self.aws_addr = '169.254.169.254'

    def _get_document(self) -> dict:
        if not self.document:
            try:
                self.document = requests.get(
                    'http://{}/latest/dynamic/instance-identity/document'.format(self.aws_addr),
                    timeout=5).json()
                _LOG.info("Amazon specific information loaded from AWS: {}".format(
                    json.dumps(self.document, indent=2)))
            except Exception as ex:
                _LOG.warn('Failed to download AWS document', exc_info=ex)
                self.document = Amazon.NONE
        return self.document if self.document != Amazon.NONE else None

    def get_aws_region(self) -> str:
        doc = self._get_document()
        return doc['region'] if doc else None

    def get_own_ip(self) -> str:
        doc = self._get_document()
        return doc['privateIp'] if doc else '127.0.0.1'

    def get_addresses_by_lb_name(self, lb_name) -> list:
        region = self.get_aws_region()

        private_ips = []

        if region is not None:
            elb = boto3.client('elb', region_name=region)
            ec2 = boto3.client('ec2', region_name=region)

            response = elb.describe_instance_health(LoadBalancerName=lb_name)

            for instance in response['InstanceStates']:
                if instance['State'] == 'InService':
                    private_ips.append(ec2.describe_instances(
                        InstanceIds=[instance['InstanceId']])['Reservations'][0]['Instances'][0]['PrivateIpAddress'])

        else:
            private_ips = [lb_name]
        _LOG.info("Ip addresses for {} are: {}".format(lb_name, private_ips))
        return private_ips



from collections import namedtuple
import logging
from operator import attrgetter

from bubuku.broker import BrokerManager
from bubuku.controller import Check
from bubuku.features.rebalance import BaseRebalanceChange
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.swap_partitions')

TpData = namedtuple('TpData', ('topic', 'partition', 'size', 'replicas'))


class SwapPartitionsChange(BaseRebalanceChange):
    def __init__(self, zk: BukuExhibitor, fat_broker_id: int, slim_broker_id: int, gap: int, size_stats: dict):
        self.zk = zk
        self.fat_broker_id = fat_broker_id
        self.slim_broker_id = slim_broker_id
        self.gap = gap
        self.size_stats = size_stats
        self.to_move = None

    def run(self, current_actions):
        if self.should_be_paused(current_actions):
            _LOG.info("Pausing swap partitions change as there are conflicting actions: {}".format(current_actions))
            return True
        try:
            _LOG.info("Running swap partitions change: {}".format(self))
            return self.__run_internal(current_actions)
        except Exception:
            _LOG.warn("Error occurred when performing partitions swap change", exc_info=True)
            return False

    def __run_internal(self, current_actions):
        # if there is already a swap which was postponed - just execute it
        if self.to_move:
            return not self.__perform_swap(self.to_move)

        # if there's a rebalance currently running - postpone current change
        if self.zk.is_rebalancing():
            return True

        # merge topics size stats to a single dict
        topics_stats = {}
        for _, broker_stats in self.size_stats.items():
            for topic in broker_stats["topics"].keys():
                if topic not in topics_stats:
                    topics_stats[topic] = {}
                topics_stats[topic].update(broker_stats["topics"][topic])

        # find partitions that are candidates to be swapped between "fat" and "slim" brokers
        swap_partition_candidates = self.__find_all_swap_candidates(self.fat_broker_id, self.slim_broker_id,
                                                                    topics_stats)

        # smallest partition from slim broker is the one we move to fat broker
        slim_broker_smallest_partition = min(swap_partition_candidates[self.slim_broker_id], key=attrgetter("size"))
        if not slim_broker_smallest_partition:
            _LOG.info("No partitions on slim broker(id: {}) found to swap".format(self.slim_broker_id))
            return False
        _LOG.info("Slim broker(id: {}) partition to swap: {}"
                  .format(self.slim_broker_id, slim_broker_smallest_partition))

        # find the best fitting fat broker partition to move to slim broker
        # (should be as much as possible closing the gap between brokers)
        fat_broker_swap_candidates = swap_partition_candidates[self.fat_broker_id]
        matching_swap_partition = self.__find_best_swap_candidate(fat_broker_swap_candidates, self.gap,
                                                                  slim_broker_smallest_partition.size)

        # if there is no possible swap that will decrease the gap - just do nothing
        if not matching_swap_partition:
            _LOG.info("No candidate from fat broker(id:{}) found to swap".format(self.fat_broker_id))
            return False
        _LOG.info("Fat broker(id: {}) partition to swap: {}".format(self.fat_broker_id, matching_swap_partition))

        # write rebalance-json to ZK; Kafka will read it and perform the partitions swap
        self.to_move = self.__create_rebalance_list(slim_broker_smallest_partition, self.slim_broker_id,
                                                    matching_swap_partition, self.fat_broker_id)
        scheduled_rebalance = self.__perform_swap(self.to_move)
        if scheduled_rebalance:
            _LOG.info("Swap partitions rebalance node was successfully created in ZK")
        else:
            _LOG.info("Swap partitions is postponed as there was a rebalance node in ZK")
        return not scheduled_rebalance

    def __perform_swap(self, rebalance_list):
        _LOG.info("Writing rebalance-json to ZK for partitions swap: {}".format(rebalance_list))
        return self.zk.reallocate_partitions(rebalance_list)

    def __find_all_swap_candidates(self, fat_broker_id: int, slim_broker_id: int, topics_stats: dict) -> dict:
        partition_assignment = self.zk.load_partition_assignment()
        swap_partition_candidates = {}
        for topic, partition, replicas in partition_assignment:
            if topic not in topics_stats or str(partition) not in topics_stats[topic]:
                continue  # we skip this partition as there is not data size stats for it

            if fat_broker_id in replicas and slim_broker_id in replicas:
                continue  # we skip this partition as it exists on both involved brokers

            for broker_id in [slim_broker_id, fat_broker_id]:
                if broker_id in replicas:
                    if broker_id not in swap_partition_candidates:
                        swap_partition_candidates[broker_id] = []
                    swap_partition_candidates[broker_id].append(
                        TpData(topic, partition, topics_stats[topic][str(partition)], replicas))
        return swap_partition_candidates

    @staticmethod
    def __find_best_swap_candidate(candidates: list, brokers_gap: int, partition_to_swap_size: int) -> TpData:
        candidates.sort(key=attrgetter("size"), reverse=True)
        matching_swap_partition = None
        smallest_new_gap = brokers_gap
        for tp in candidates:
            new_gap = abs(brokers_gap - 2 * abs(tp.size - partition_to_swap_size))
            if new_gap < smallest_new_gap:
                smallest_new_gap = new_gap
                matching_swap_partition = tp
        return matching_swap_partition

    def __create_rebalance_list(self, tp1: TpData, br1: int, tp2: TpData, br2: int) -> list:
        return [
            (tp1.topic, tp1.partition, self.__replace_broker(tp1.replicas, br1, br2, tp2.replicas[0] == br2)),
            (tp2.topic, tp2.partition, self.__replace_broker(tp2.replicas, br2, br1, tp1.replicas[0] == br1))
        ]

    def __replace_broker(self, replicas: list, broker_to_replace: int, replacement: int, was_leader: bool) -> list:
        rps = [x for x in replicas if x != broker_to_replace]
        if was_leader:
            return [replacement] + rps
        else:
            return rps + [replacement]

    def __str__(self):
        return 'SwapPartitions(fat_broker_id: {}, slim_broker_id: {}, gap: {})'.format(
            self.fat_broker_id, self.slim_broker_id, self.gap)


class CheckBrokersDiskImbalance(Check):
    def __init__(self, zk: BukuExhibitor, broker: BrokerManager, diff_threshold_kb: int):
        super().__init__(check_interval_s=900)
        self.zk = zk
        self.broker = broker
        self.diff_threshold_kb = diff_threshold_kb

    def check(self):
        if self.broker.is_running_and_registered():
            _LOG.info("Starting broker disk imbalance check")
            try:
                return self.create_swap_partition_change()
            except Exception:
                _LOG.warn("Error occurred when performing disk imbalance check", exc_info=True)
        return None

    def create_swap_partition_change(self) -> SwapPartitionsChange:
        size_stats = self.zk.get_disk_stats()
        if not size_stats or len(size_stats.keys()) == 0:
            _LOG.info("No size stats available, imbalance check cancelled")
            return None

        # find the most "slim" and the most "fat" brokers
        def free_size_getter(tup):
            return tup[1]["disk"]["free_kb"]

        slim_broker_id = max((item for item in size_stats.items()), key=free_size_getter)[0]
        fat_broker_id = min((item for item in size_stats.items()), key=free_size_getter)[0]
        fat_broker_free_kb = size_stats[fat_broker_id]["disk"]["free_kb"]
        slim_broker_free_kb = size_stats[slim_broker_id]["disk"]["free_kb"]

        # is the gap is big enough to swap partitions?
        gap = slim_broker_free_kb - fat_broker_free_kb
        if gap < self.diff_threshold_kb:
            _LOG.info("Gap between brokers: (id: {}, free_kb: {}) and (id: {}, free_kb: {}) is not enough to "
                      "trigger partitions swap; gap is {} KB".format(slim_broker_id, slim_broker_free_kb, fat_broker_id,
                                                                     fat_broker_free_kb, gap))
            return None
        else:
            _LOG.info("Creating swap partitions change to swap partitions of brokers (id: {}, free_kb: {}) and "
                      "(id: {}, free_kb: {}); gap is {} KB".format(slim_broker_id, slim_broker_free_kb, fat_broker_id,
                                                                   fat_broker_free_kb, gap))
            return SwapPartitionsChange(self.zk, int(fat_broker_id), int(slim_broker_id), gap, size_stats)



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Check, Change
from bubuku.features.rebalance import RebalanceChange
from bubuku.features.restart_on_zk_change import RestartBrokerChange
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.remote_exec')


class RemoteCommandExecutorCheck(Check):
    def __init__(self, zk: BukuExhibitor, broker_manager: BrokerManager):
        super().__init__()
        self.zk = zk
        self.broker_manager = broker_manager

    def check(self) -> Change:
        with self.zk.lock():
            data = self.zk.take_action(self.broker_manager.id_manager.get_broker_id())
        if not data:
            return None
        if 'name' not in data:
            _LOG.error('Action name can not be restored from {}, skipping'.format(data))
            return None
        if data['name'] == 'restart':
            return RestartBrokerChange(self.zk, self.broker_manager, lambda: False)
        if data['name'] == 'rebalance':
            return RebalanceChange(self.zk, self.zk.get_broker_ids())
        return None

    @staticmethod
    def register_restart(zk: BukuExhibitor, broker_id: str):
        with zk.lock():
            zk.register_action(
                {'name': 'restart'},
                broker_id=broker_id)

    @staticmethod
    def register_rebalance(zk: BukuExhibitor, broker_id: str):
        with zk.lock():
            if broker_id:
                zk.register_action({'name': 'rebalance'}, broker_id=broker_id)
            else:
                zk.register_action({'name': 'rebalance'})
